# generated from codegen/templates/_boundingbox3d.py

from __future__ import annotations

__all__ = ["DBoundingBox3d", "DBoundingBox3dOverlappable", "HasDBoundingBox3d"]

from typing import TYPE_CHECKING
from typing import Iterable
from typing import Protocol
from typing import overload

from emath import DVector3

from ._separating_axis_theorem import separating_axis_theorem

if TYPE_CHECKING:
    from ._drectanglefrustum import DRectangleFrustum


class DBoundingBox3dOverlappable(Protocol):
    def overlaps_d_bounding_box_3d(self, other: DBoundingBox3d) -> bool: ...


class HasDBoundingBox3d(Protocol):
    @property
    def bounding_box(self) -> DBoundingBox3d: ...


class DBoundingBox3d:
    __slots__ = ["_extent", "_position", "_size"]

    @overload
    def __init__(self, position: DVector3, size: DVector3) -> None: ...

    @overload
    def __init__(self, *, shapes: Iterable[HasDBoundingBox3d | DVector3]) -> None: ...

    def __init__(
        self,
        position: DVector3 | None = None,
        size: DVector3 | None = None,
        *,
        shapes: Iterable[HasDBoundingBox3d | DVector3] | None = None,
    ):
        if shapes is not None:
            if position is not None:
                raise TypeError("position cannot be supplied with shapes argument")
            if size is not None:
                raise TypeError("size cannot be supplied with shapes argument")
            accum_position: DVector3 | None = None
            accum_extent: DVector3 | None = None
            for s in shapes:
                if isinstance(s, DVector3):
                    p = e = s
                else:
                    p = s.bounding_box.position
                    e = s.bounding_box.extent
                if accum_position is None:
                    accum_position = p
                else:
                    accum_position = DVector3(
                        min(p.x, accum_position.x),
                        min(p.y, accum_position.y),
                        min(p.z, accum_position.z),
                    )
                if accum_extent is None:
                    accum_extent = e
                else:
                    accum_extent = DVector3(
                        max(e.x, accum_extent.x),
                        max(e.y, accum_extent.y),
                        max(e.z, accum_extent.z),
                    )
            if accum_position is None:
                position = DVector3(0)
                size = DVector3(0)
            else:
                assert accum_extent is not None
                position = accum_position
                size = accum_extent - accum_position

        assert position is not None
        assert size is not None
        if size < DVector3(0):
            raise ValueError("each size dimension must be >= 0")
        self._position = position
        self._size = size
        self._extent = self._position + self._size

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DBoundingBox3d):
            return False
        return self._position == other._position and self._size == other._size

    def __repr__(self) -> str:
        return f"<BoundingBox3d position={self._position} size={self._size}>"

    def overlaps(self, other: DVector3 | DBoundingBox3dOverlappable) -> bool:
        if isinstance(other, DVector3):
            return self.overlaps_d_vector_3(other)
        try:
            other_overlaps = other.overlaps_d_bounding_box_3d
        except AttributeError:
            raise TypeError(other)
        return other_overlaps(self)

    def overlaps_d_rectangle_frustum(self, other: DRectangleFrustum) -> bool:
        return separating_axis_theorem(
            {
                DVector3(1, 0, 0),
                DVector3(0, 1, 0),
                DVector3(0, 0, 1),
                *(p.normal for p in other.planes),
            },
            set(other.points),
            set(self.points),
        )

    def overlaps_d_bounding_box_3d(self, other: DBoundingBox3d) -> bool:
        return not (
            self._position.x >= other._extent.x
            or self._extent.x <= other._position.x
            or self._position.y >= other._extent.y
            or self._extent.y <= other._position.y
            or self._position.z >= other._extent.z
            or self._extent.z <= other._position.z
        )

    def overlaps_d_vector_3(self, other: DVector3) -> bool:
        return (
            other.x >= self._position.x
            and other.x < self._extent.x
            and other.y >= self._position.y
            and other.y < self._extent.y
            and other.z >= self._position.z
            and other.z < self._extent.z
        )

    def translate(self, translation: DVector3) -> DBoundingBox3d:
        return DBoundingBox3d(self._position + translation, self._size)

    @property
    def bounding_box(self) -> DBoundingBox3d:
        return self

    @property
    def extent(self) -> DVector3:
        return self._extent

    @property
    def position(self) -> DVector3:
        return self._position

    @property
    def size(self) -> DVector3:
        return self._size

    @property
    def points(
        self,
    ) -> tuple[DVector3, DVector3, DVector3, DVector3, DVector3, DVector3, DVector3, DVector3]:
        return (
            self._position,
            self._position + self._size.xoo,
            self._position + self._size.oyo,
            self._position + self._size.ooz,
            self._position + self._size.xyo,
            self._position + self._size.xoz,
            self._position + self._size.oyz,
            self._extent,
        )
